#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>
#include "poly.h"
#include "basic.h"

/*******************/
/* Concurrent FIFO */
/*******************/

typedef struct {
  pthread_mutex_t *mutex ;
  pthread_cond_t *is_empty, *is_full ;

  int sz ;
  int fst,lst,nitems ;
  subtask_t **t ;
} fifo_t ;

fifo_t *alloc_fifo(int sz) {
  fifo_t *r = malloc_check(sizeof(*r)) ;
  r->fst = r->lst = r->nitems = 0 ;
  r->sz = sz ;
  r->t = calloc(sz,sizeof(*r->t)) ;
  if (!r->t) {
    perror("calloc") ;
    exit(2) ;
  }
  r->mutex = alloc_mutex() ;
  r->is_empty = alloc_cond() ;
  r->is_full = alloc_cond() ;
  return r ;
}

void free_fifo(fifo_t *f) {
  free_cond(f->is_empty) ;
  free_cond(f->is_full) ;
  free_mutex(f->mutex) ;
  free(f->t) ;
  free(f) ;
}

/* real operations */

void put(fifo_t *f,subtask_t *z) {
  lock_mutex(f->mutex) ;
  while (f->nitems == f->sz) {
    wait_cond(f->is_full,f->mutex) ;
  }
  int was_empty = f->nitems == 0 ;
  f->t[f->lst] = z ;
  f->lst++ ; f->lst %= f->sz ; f->nitems++ ;
  if (was_empty) {
    signal_cond(f->is_empty) ;
  }
  unlock_mutex(f->mutex) ;
}

subtask_t *get(fifo_t *f) {
  subtask_t *r ;
  lock_mutex(f->mutex) ;
  while (f->nitems == 0) {
    wait_cond(f->is_empty,f->mutex) ;
  }
  int was_full = f->nitems == f->sz ;
  r = f->t[f->fst] ;
  f->fst++ ; f->fst %= f->sz ; f->nitems-- ;
  if (was_full) {
    signal_cond(f->is_full) ;
  }
  unlock_mutex(f->mutex) ;
  return r ;
}


/****************/
/* Main control */
/****************/
typedef struct {
  int id ;
} ctx_t ;
static fifo_t *fifo ;
static pthread_mutex_t *mutex ;
static volatile count_t sum ;

void *slave(void *p) {
  int id = ((ctx_t *)p)->id ;
  for ( ; ; ) {
    subtask_t *z = get(fifo) ;
    if (z == NULL) {
      put(fifo,z) ;
      if (verbose)  fprintf(stderr,"proc %i over\n",id) ;
      return NULL ;
    }
    if (verbose) fprintf(stderr,"proc %i got task %i\n",id,z->id) ;
    count_t c = run_subtask(z) ;
    if (verbose) fprintf(stderr,"proc %i returns task %i\n",id,z->id) ;
    lock_mutex(mutex) ;
    sum += c ;
    unlock_mutex(mutex) ;
  }
}

/* Generator */
void emit_fifo(subtask_t *z) {
  put(fifo,copy_arg(z)) ;
}

count_t run(int n, int depth, int nprocs, int fsz) {
  /* Initialise */
  sum = 0 ;
  fifo = alloc_fifo(fsz) ;
  mutex = alloc_mutex() ;
  
  /* Start workers */
  pthread_t th[nprocs] ;
  ctx_t ctx[nprocs] ;
  for (int k = 0 ; k < nprocs ; k++) {
    ctx_t *p = &ctx[k] ;
    p->id = k ;
    create_thread(&th[k],slave,p) ;
  }

  /* Fill up fifo */
  int ntasks = generate_subtasks(n, depth, emit_fifo) ;
  if (verbose) {
    fprintf(stderr,"Subtasks: %i\n",ntasks) ;
  }
  put(fifo,NULL) ;
  for (int k = 0 ; k < nprocs ; k++) {
    (void)join_thread(&th[k]) ;
  }

  free_fifo(fifo) ;
  free_mutex(mutex) ;
  return sum ;
}
void usage(char *pgm)
{
  fprintf(stderr,"usage: %s [-v]*\n",pgm);
  exit(-1);
}

int main(int argc, char **argv) {
  char *arg=0;
  char *pgm = argv[0];
  int n = 20 ;
  int gen_now = 15 ;
  int nprocs = 4 ;
  int fsz = -1 ;

  while (--argc) {
    arg = *++argv;
    if (*arg == '-') {
      arg++ ;
       if (*arg == 'g') {
        if (arg[1]) {
          gen_now += atoi(arg+1) ;
        } else {
          usage(pgm) ;
        }
       } else if (*arg == 'p' || *arg == 'j') {
         if (arg[1]) {
           nprocs = atoi(arg+1) ;
         } else {
           usage(pgm) ;
         }
       } else  if (*arg == 'f') {
         if (arg[1]) {
           fsz = atoi(arg+1) ;
         } else {
           usage(pgm) ;
         }
       } else if (*arg == 'v') {
         verbose++;
       } else {
         usage(pgm);
       }
    } else break ;      
  }
  if (argc >= 1) n = atoi(arg);
  if (n > NMAX) {
      fprintf(stderr,"argument must be <= %d\n",NMAX);
      exit(2);
  }
  if (fsz <= 0) fsz = nprocs+1 ;

  count_t r = run(n,gen_now,nprocs,fsz) ;
  printf("%" PRC "\n",r) ;
  return 0 ;
}
